/*
 * MVME16x loader module.  Designed to be patched into the ELF
 * header of the vmlinux file at offset 0x800.
 *
 * Richard Hirst    21-10-96
 */

#define noDEBUG

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>
#include <asm/byteorder.h>
#include <linux/elf.h>

typedef unsigned char	uchar;

#define SAVE_HDRS	0x400	/* Offset to section header save area */
#define BI_START	0x100	/* boot_info starts off here */
#define BI_LEN		0x100	/* Guess - should be plenty */

#ifndef DEBUG
#define base	0		/* In real life it always runs at zero */
#endif

void fill (void *dst, int len);
void mv (void *dst, void *src, int len);
ulong locate (ulong);




int main (int argc, char **argv)
{
	uchar *hp;		/* Current start of vmlinux image */
	Elf32_Ehdr *ehp;	/* ELF header struct (=hp) */
	Elf32_Shdr *esp;	/* ELF section header pointer */
	int i;
	ulong bi;		/* Destination for boot_info */
#ifdef DEBUG
	Elf32_Shdr *shp;	/* shstrtab section header pointer */
	char *ssp;		/* shstrtab start address */
	int ifd, len;		/* Input file descriptor, file length */
	ulong base;		/* Actual base of malloc-ed memory image */
#else
	/*
	 * A function pointer, initialised to 0x1000, which is the
	 * entry address in head.S.
	 */
	void (* goforit)(void) = (void (*)(void))4096;
#endif


#ifdef DEBUG
	/* Open the input file, read into malloc-ed memory */
	ifd = open (*++argv, O_RDONLY);
	if (ifd < 0)
	{
		perror ("open");
		exit(1);
	}
	hp = malloc (0x200000);
	if (hp < 0)
	{
		perror ("malloc");
		exit (1);
	}
	/* Force hp to be page aligned */
	hp = (uchar *)(((ulong)hp + 0xfff) & ~0xfff);
	base = (ulong)hp;
	if ((len = read (ifd, hp, 0x200000)) <= 0)
	{
		perror ("read");
		exit (1);
	}
#else
	/*
	 * Header is at start of image.  We are currently somewhere in
	 * the first 4K page....
	 */
	/* hp = (uchar *)(((ulong)main + 0xfff) & ~0xfff); */
	hp = (uchar *)locate(0);
#endif

	ehp = (Elf32_Ehdr *)hp;

	/* Copy section headers to somewhere safe */

	mv (hp + SAVE_HDRS, hp + ntohl(ehp->e_shoff), ntohs(ehp->e_shnum) * ntohs(ehp->e_shentsize));

#ifdef DEBUG
	/* Display segment header table */
	esp = (Elf32_Shdr *)(hp + SAVE_HDRS);

	shp = (Elf32_Shdr *)(hp + SAVE_HDRS + ntohs(ehp->e_shstrndx) * ntohs(ehp->e_shentsize));
	ssp = (char *)(hp + htonl(shp->sh_offset));

	printf ("%d section headers at offset 0x%08lx\n", ntohs(ehp->e_shnum), ntohl(ehp->e_shoff));

	for (i = 0; i < ntohs(ehp->e_shnum); i++)
	{
		printf ("%2d: %12s offset=%08lx addr=%08lx size=%08lx\n", i, ssp + htonl(esp->sh_name),
			htonl(esp->sh_offset), htonl(esp->sh_addr), htonl(esp->sh_size));
		esp = (Elf32_Shdr *)((ulong)esp + ntohs(ehp->e_shentsize));
	}
#endif

	/*
	 * Work from last segment down, copying or zero filling as required.
	 * Assumes segments are in ascending address order, and that any which
	 * need to move need to move to higher addresses.
	 * If the image was loaded somewhere other than zero (eg at 0x200000)
	 * then ordering and overlap wont matter.
	 */

	for (i = ntohs(ehp->e_shnum) - 1; i >= 0; i--)
	{
		esp = (Elf32_Shdr *)(hp + SAVE_HDRS + ntohs(ehp->e_shentsize) * i);
		if (htonl(esp->sh_addr) && (htonl(esp->sh_addr) + base) != (htonl(esp->sh_offset) + (ulong)hp))
		{
			if (htonl(esp->sh_type) == SHT_NOBITS)
			{
				/* Need to zero fill */
#ifdef DEBUG
				printf ("Need to fill from %08lx for %08lx bytes\n",
					htonl(esp->sh_addr), htonl(esp->sh_size));
#endif
				fill ((void *)htonl(esp->sh_addr) + base, htonl(esp->sh_size));

				/* This must be .bss - record where boot_info need placing */
				bi = htonl(esp->sh_addr) + htonl(esp->sh_size);
			}
			else
			{
				/* Need to move it */
#ifdef DEBUG
				printf ("Need to move from %08lx to %08lx, %08lx bytes\n",
					htonl(esp->sh_offset), htonl(esp->sh_addr), htonl(esp->sh_size));
#endif
				mv ((void *)htonl(esp->sh_addr) + base, htonl(esp->sh_offset) + hp, htonl(esp->sh_size));
			}
		}
	}
	/*
	 * Now we need to position the boot_info structure at the end of bss.
	 */
	mv ((void *)bi + base, BI_START + hp, BI_LEN);
#ifdef DEBUG
	printf ("boot_info copied to base + %08lx\n", bi);
#endif

#ifndef DEBUG
	goforit();	/* Go enter the real code head.S */
#endif
	return 0;
}


/* Simple top-down memcpy */

void mv (void *dst, void *src, int len)
{
	volatile ulong *d = (ulong *)dst;
	volatile ulong *s = (ulong *)src;

	len = (len + 3) >> 2;
	d += len;
	s += len;
	while (len--)
		*--d = *--s;
}

/* Simple fill */

void fill (void *dst, int len)
{
	volatile ulong *p = (ulong *)dst;

	len = (len + 3) >> 2;
	while (len--)
		*p++ = 0;
}


ulong locate (ulong n)
{
	return (*(&n-1) & ~0xfff);
}

